from ..materials import *
from .sparams import *
from .setup import *
from ..constants import *
from ..layers import *
from ..utils import *
import gdsfactory as gf
from copy import deepcopy
from functools import partial
from math import cos, pi, sin
import os
from gdsfactory.generic_tech import LAYER_STACK, LAYER
import json


def make_pic_inv_problem(path, c,  targets, iters=10,
                         lvoid=0, lsolid=0, symmetries=[],
                         weights=dict(),
                         eta=.4, init=1,   stoploss=None,
                         design_region_layer=DESIGN_LAYER,
                         #    design_guess_layer=LAYER.GUESS,
                         fill_layer=LAYER.WG,
                         void_layer=None,
                         layer_stack=LAYER_STACK,
                         restart=True, save_memory=False, **kwargs):
    design_region_layer = tuple(design_region_layer)
    # if not N:
    #     raise NotImplementedError(
    #         "3D inverse design feature must be requested from Luminescent AI info@luminescentai.com")

    if "phase_shifter" in targets:
        keys = ["o2@0,o1@0"]
        wavelengths = [targets["phase_shifter"]]
    else:
        if "tparams" in targets:
            for center_wavelength in targets["tparams"]:
                d = {}
                # for k in targets["tparams"][center_wavelength]:
                #     po, mo, pi, mi = unpack_sparam_key(k)
                #     # if mo == 0:
                #     #     _mo = 1
                #     # elif mo == 1:
                #     #     _mo = 0
                #     # else:
                #     #     mo = None

                #     # if mo is not None:
                #     #     _k = f"{po}@{_mo},{pi}@{mi}"
                #     #     if _k not in targets["tparams"][center_wavelength]:
                #     #         d[_k] = 0

                #     _k = f"{pi}@{mi},{pi}@{mi}"
                #     if _k not in targets["tparams"][center_wavelength]:
                #         d[_k] = 0
                #     # _k = f"o{pi}@1,o{pi}@{mi}"
                #     # if _k not in targets["tparams"][center_wavelength]:
                #     #     d[_k] = 0
                # targets["tparams"][center_wavelength].update(d)
        print(targets)

        targets1 = {}
        keys = SortedSet()
        wavelengths = SortedSet()
        for k, d in targets.items():
            d = {
                center_wavelength: {
                    long_sparam_key(k): v for k, v in d.items()
                } for center_wavelength, d in d.items()}
            targets1[k] = d
            if k in ["sparams", "tparams"]:
                for s in sum([list(d.keys()) for d in d.values()], []):
                    keys.add(s)
            for center_wavelength in d:
                wavelengths.add(center_wavelength)
        targets = targets1
        keys = list(keys)
        print(keys)
        wavelengths = list(wavelengths)

    wavelengths0 = deepcopy(wavelengths)
    prob = make_pic_sim_problem(path, c,
                                layer_stack=layer_stack,
                                wavelengths=wavelengths,
                                study="inverse_design",
                                keys=keys, ** kwargs)
    wavelengths = prob["wavelengths"]

    prob["restart"] = restart
    prob["weights"] = {**{
        "tparams": 1,
        "sparams": 1,
        "phasediff": 1,
    }, **weights}
    prob["save_memory"] = save_memory

    _targets = {}
    for k in targets:
        if k in ["sparams", "tparams"]:
            _targets[k] = {}
            for (w, w0) in zip(wavelengths, wavelengths0):
                _targets[k][w] = targets[k][w0]
    targets = _targets

    prob["targets"] = targets
    prob["wavelengths"] = wavelengths
    # prob["init"] = init
    prob["eta"] = eta
    polys = c.extract([design_region_layer]).get_polygons()
    if not symmetries:
        symmetries = [[]]*len(polys)
    else:
        if type(symmetries[0]) is not list:
            symmetries = [symmetries]*len(polys)

    def _bbox(b):
        return [[b.left/1e3, b.bottom/1e3], [b.right/1e3, b.top/1e3]]
    prob["designs"] = [
        {
            "layer": design_region_layer,
            "bbox": _bbox(p.bbox()),
            "symmetries": s,
            "init": init,
        } for p, s in zip(list(polys.values())[0], symmetries)
    ]
    prob["lvoid"] = lvoid
    prob["lsolid"] = lsolid
    prob["stoploss"] = stoploss
    prob["design_config"] = dict()
    l = get_layers(layer_stack, fill_layer)[0]
    d = {"thickness": l.thickness,
         "material": matname(l.material), "zmin": l.zmin}
    d["layer"] = fill_layer
    prob["design_config"]["fill"] = d

    # if void_layer is not None:
    #     l = get_layers(layer_stack, void_layer)[0]
    #     d = {"thickness": l.thickness, "material": l.material, "zmin": l.zmin}
    #     d["layer"] = void_layer
    #     d["epsilon"] = materials[d["material"]]["epsilon"]
    #     d["epsilon"] = d["epsilon"]
    # else:
    #     d = copy.deepcopy(d)
    #     d["epsilon"] = epsmin
    #     d["epsilon"] = d["epsilon"]
    # prob["design_config"]["void"] = d

    prob["design_config"]["design_region_layer"] = design_region_layer
    prob["iters"] = iters
    save_problem(prob, path)
    return prob


def apply_design(c0,  sol):
    path = sol["path"]
    a = gf.Component()
    a.add_ref(c0)
    fill = sol["design_config"]["fill"]["layer"]
    dl = sol["design_config"]["design_region_layer"]
    for i, d in enumerate(sol["designs"]):
        x0, y0 = d["bbox"][0]
        x1, y1 = d["bbox"][1]
        b = gf.Component()
        b.add_polygon(
            [(x0, y0), (x1, y0), (x1, y1), (x0, y1)], layer=fill)
        a = gf.boolean(a, b, "not", layer=fill)
    c = gf.Component()
    c << a
    # for layer in c0.layers:
    #     if layer != dl:
    #         c.add_ref(c0.extract([layer]))
    # c.show()
    # raise ValueError()
    g = gf.import_gds(os.path.join(path, f"optimized_design_region_{i+1}.gds"))
    polygons = g.get_polygons(merge=True)
    g = gf.Component()
    for p in polygons[1]:
        g.add_polygon(p, layer=fill)
    g = c << g
    g.xmin = x0
    g.ymin = y0
    return c



import gdsfactory as gf
from layers import *
core_layer = (1, 0)


def phase_shifter(l, wwg, nbends, ubend):
    c = gf.Component()
    ls = l-2*ubend.size[0]
    s0 = c << gf.path.extrude(gf.path.straight(
        length=ls), width=wwg, layer=core_layer)
    s = s0

    lu = ubend.size[0]
    wdope = ubend.size[1]-wwg
    sin = c << gf.path.extrude(gf.path.straight(
        length=lu), width=wwg, layer=core_layer)
    sin.connect("o2", s.ports["o1"])

    for i in range(nbends+2):
        if 0 < i < nbends+1:
            u_ = (c << ubend)
            p1 = "o1" if i % 2 == 1 else "o2"
            p2 = "o2" if i % 2 == 1 else "o1"
            u_.connect(p1, s.ports["o2"])
            s_ = (c << gf.path.extrude(gf.path.straight(
                length=ls), width=wwg, layer=core_layer))
            s_.connect("o1", u_.ports[p2])
            s = s_

        pn = "p" if i % 2 == 0 else "n"
        pn_layer = p_layer if pn == "p" else n_layer
        label = "p doping" if pn == "p" else "n doping"

        pn = c << gf.components.rectangle(size=(l, wdope), layer=pn_layer)
        pn.xmin = c.xmin
        if i < nbends+1:
            pn.ymin = s.center[1]
        else:
            pn.ymax = s.center[1]
        c.add_label(label, position=pn.center, layer=(501, 0))

    sout = c << gf.path.extrude(gf.path.straight(
        length=lu), width=wwg, layer=core_layer)
    sout.connect("o1", s.ports["o2"])

    heater_margin = 0.1
    center = c.center
    heater = c << gf.components.rectangle(
        size=(l+2*heater_margin, c.size[1]+2*heater_margin), layer=(61, 0))
    heater.center = center
    c.add_label("heater pad", position=heater.bbox_np()[1], layer=(502, 0))

    c.add_label("zigzag phase shifter", position=(
        c.center[0], c.ymax), layer=(500, 0))
    c.show()
    return c


def ubend_resonator(l, wwg, ubend):
    c = gf.Component()
    arm = gf.Component()
    quarter = gf.Component()

    r = 1
    common_left_pad_amount = 8
    lsb = 6
    ls = (l-common_left_pad_amount-2*lsb-2*ubend.size[0])/2

    su = quarter << gf.components.straight(
        length=ls, width=wwg, layer=core_layer)  # Straight section
    sc = quarter << gf.components.straight(
        length=common_left_pad_amount/2, width=wwg, layer=core_layer)  # Straight section
    sb = quarter << gf.components.bend_s(
        (lsb, 1), 32, gf.cross_section.strip(wwg), )
    sb.mirror(p1=(0, 0), p2=(1, 0))
    sb.connect("o2", sc.ports["o1"])
    su.connect("o2", sb.ports["o1"])
    quarter.add_port("o1", port=su.ports["o1"])
    quarter.add_port("o2", port=sc.ports["o2"])

    coupler_margin = .15

    sw = arm << quarter
    sw.ymin = +wwg/2+coupler_margin
    sw.xmax = 0

    se = arm << quarter
    se = se.mirror()
    se.connect("o2", sw.ports["o2"])

    arm.add_port("o1", port=sw.ports["o1"])
    arm.add_port("o2", port=se.ports["o1"])
    south = c << arm

    ubend1 = c << ubend
    ubend2 = c << ubend
    ubend1.connect("o1", south.ports["o1"])
    ubend2.connect("o2", south.ports["o2"])

    north = (c << arm)
    # north.mirror(p1=(0,0),p2=(1,0))
    north.connect("o2", ubend1.ports["o2"])

    xmin = c.xmin
    ymin = c.ymin
    ymax = c.ymax

    leg1 = c << gf.components.straight(length=l, width=wwg, layer=core_layer)
    # c.add_port(name="o1", center=(0, 0), width=wwg, orientation=0)
    leg1.ymax = ymin-coupler_margin
    leg1.xmin = xmin

    leg2 = c << gf.components.straight(length=l, width=wwg, layer=core_layer)
    leg2.ymin = ymax+coupler_margin
    leg2.xmin = xmin

    c.add_label("zigzag resonator", position=c.center)
    c.show()
    return c
